#include <math.h>
#include <stdio.h>
#include "physics.h"
#include "keycvt.h"
#include "vop.h"
#include "chn.h"
#include "Malloc.h"
#include "Msgs.h"
#include "error.h"
#include "Assert.h"
#include "fastflpt.h"
#include "keycvt.h"
#include "protos.h"

#include "tree.h"
#include "decomp.h"

static size_t
GravMakeIntermediate(Key_t key, int ntbody, 
		     const tbody **frombodies, int ncell,
		     const cofm **fromcells, cofm *to){
  int i;
  const tbody *bp;
  float dmass;
  float newbmax;
  float center[NDIM], cellsz[NDIM];
  Vxd(float dx);
  tbbox CellBbox;

  to->mass = 0.;
  VS(to->pos, = 0.);
  to->B2 = 0.;
  to->bmax = 0.;
  to->daughters = 0;
  to->cost = 0.F;
  /* Get the cm of the new cell. */
  /* First loop over daughter cells */
  Msgf(("Aggravating %d cells\n", ncell));
  for(i=0; i<ncell; i++){
    const cofm *dp = fromcells[i];
    /* If this daughter doesn't exist, bail out now. */
    assert(dp);
    to->daughters += dp->daughters;
    dmass = dp->mass;
    to->mass += dmass;
    to->cost += dp->cost;
    Msgf(("\tdaughter cell %d @%p: mass: %g, " Sinfix("%.2f", " ") "\n",
	  i, dp, dp->mass, Vinfix(dp->pos, COMMA)));
    VV(to->pos, += dmass * dp->pos);
  }

  /* And then over the terminal bodies in the cell */
  Msgf(("Aggravating %d bodies\n", ntbody));
  for(i=0; i<ntbody; i++){
    bp = frombodies[i];
    assert(bp);
    to->daughters++;
    dmass = bp->mass;
    Msgf(("\tdaughter body %d @%p: mass: %g, " Sinfix("%.2f", " ") "\n",
	  i, bp, bp->mass, Vinfix(bp->pos, COMMA)));
    to->mass += dmass;
    VV(to->pos, += dmass * bp->pos);
    to->cost += bp->cost;
  }

  /* Divide out the total mass */
  if( to->mass != (float)0. ){
    float massinv = recipf(to->mass);
    VS(to->pos, *= massinv);
  }else{
    SeriousWarning("Zero mass in AggregateGrav!\n");
  }

  /* Now loop again to pick up B2, etc.  */
  for(i=0; i<ncell; i++){
    float tmp[NDIM];
    float tmpsq;

    const cofm *dp = fromcells[i];
    VVV(tmp, = to->pos, - dp->pos);
    dmass = dp->mass;
    newbmax = dp->bmax;
    to->B2 += dp->B2;

    tmpsq = Dot(tmp, tmp);
    if( tmpsq != 0.F ){
      /* avoid doing a sqrtf_fast(0).  and don't bother multiplying
	 by and adding zero either */
      to->B2 += dmass * tmpsq;
      newbmax += sqrtf_fast(tmpsq);
    }
    if( newbmax > to->bmax )
      to->bmax = newbmax;
  }

  newbmax = 0.;
  for(i=0; i<ntbody; i++){
    float tmp[NDIM];
    float tmpsq;
    bp = frombodies[i];
    assert(bp);
    VVV(tmp, = to->pos, - bp->pos);
    dmass = bp->mass;
    tmpsq = Dot(tmp, tmp);
    if( tmpsq != 0.F ){
      /* avoid doing a sqrtf_fast(0).  and don't bother multiplying
	 by and adding zero either */
      to->B2 += dmass * tmpsq;
      newbmax = sqrtf_fast(tmpsq);
      if( newbmax > to->bmax )
	to->bmax = newbmax;
    }
  }

  /* This is an alternative bound on bmax, which is sometimes tighter */
  /* than the cumulative bound computed above. */
  CellBBFromKey(key, &UniverseBbox, &CellBbox, key_order);
  CenterBbox(&CellBbox, center, cellsz);
  VxVVVS(dx, = cellsz, + fabs LPAREN to->pos, - center,  RPAREN);
  newbmax = sqrtf_fast(Dotx(dx, dx));
  to->bmax = ( newbmax < to->bmax ) ? newbmax : to->bmax;
  return sizeof(cell);
}

static double a[6];		/* coef of error poly */
static void rcrit_poly(double r, double *value, double *deriv);
static double rtnewt(void (*funcd)(double, double *, double *),  
		    double x1, double xacc);

/* Turn the ptr from a cofm to a cell. */
static void *
GravMakeCell(cell *to, const cofm *from, size_t sz){
  float B2, B3;
  float bmaxhalf, rcritmax;
  float massinv;

  assert(sz == sizeof(cell));
  B2 = from->B2;

  if( B2 == (float)0. ){
    SeriousWarning("B2=0: ?\n"); /* q->rcrit = q->bmax; */
  }else if(errtol <= 0.F ){
    to->rcrit = 1.0e19;		/* rcrit*rcrit shouldn't overflow... */
  }else{
    bmaxhalf = from->bmax * (float)0.5;
    rcritmax = bmaxhalf + sqrtf_fast(bmaxhalf*bmaxhalf + sqrtf_fast(3.0F*B2/errtol));
    if (!finite(rcritmax))
      Error("Bad rcritmax, q->bmax = %g, B2 = %g\n", from->bmax, B2);
    massinv = recipf(from->mass);
    B3 = B2 * sqrtf_fast(B2*massinv); 
    to->B3 = (float)2.*B3;
    if (!finite(B2) || !finite(B3) || !finite(from->bmax))
      Error("Bad value B2 = %g, B3 = %g, bmax = %g\n", B2, B3, from->bmax);
    a[0] = 2.*B3;
    a[1] = -3. * B2;
    a[2] = 0.;
    a[3] = errtol*from->bmax*from->bmax;
    a[4] = -2. * errtol *from->bmax;
    a[5] = errtol;
    to->rcrit =  rtnewt(rcrit_poly, rcritmax, .01*rcritmax);
    to->rcrit += 0.01*rcritmax;
  }
  to->B2 = (float)3.*B2;

  to->mass = from->mass;
  VV(to->pos, = from->pos);
  to->daughters = from->daughters;
  to->bmax = from->bmax;
  to->cost = from->cost;
  Msgf(("Cell: %s\n", PrintCellContents(to)));
  return to;
}

BuildFuncs_t GravBfuncs = {
  GravMakeIntermediate, 
  GravMakeCell, 
  sizeof(cofm)};

/* This can be used for internal checking in tree17.c, i.e., set:
   tree17keyFromTbody = &keyFromTbody
   It's also used by the decomposition routines. 
*/
Key_t keyFromTbody(const _tbody *_tb){
  const tbody *tb = (const tbody *)_tb;
  Key_t answer;

  answer = KeyFromFloats(tb->pos, &UniverseBbox, nbodybits, key_order);
  return answer;
}

/* Four functions used as callbacks in decomposition and splitting */
double
cellCost(Key_t k, tcell *cp){
  return (Contents(cp))->cost;
}

void
bodyCostVisit(sib_group *sg, SplitTbodyInfo *tbinfo){
  int i;
  tbody *tb = Tbodies(sg);

  for(i=0; i<sg->ntbody; i++){
    tbinfo[i].key = keyFromTbody(tb);
    tbinfo[i].value = tb->cost;
    tb++;
  }
}

double
cellSz(Key_t k, tcell *tc){
  return (Contents(tc))->daughters;
}

void bodySzs(sib_group *sg, SplitTbodyInfo *tbinfo){
  int i;
  tbody *tb;

  tb = Tbodies(sg);
  for(i=0; i<sg->ntbody; i++){
    tbinfo[i].key = keyFromTbody(tb);
    tbinfo[i].value = 1.0;
    tb++;
  }
}

/* Use doubles here to avoid catastrophe from roundoff. */
static void 
rcrit_poly(double r, double *value, double *deriv){
    /* Do we care that a[2] and da[1] are zero? */
    double p = a[5];
    double dp = 0.;
    int n = 5;

    /* See pg. 149 of Numerical Rec. */
    /* We could unroll it... */
    while(n>0) { dp = dp*r + p; p = p*r + a[--n]; }
    if (!finite(p) || !finite(dp))
      Error("Bad p or dp, p = %g, dp = %g, n = %d\n", p, dp, n);
    *value = p;
    *deriv = dp;
}

/* From Numerical Recipes, rtnewt.c (modified) */
/* It's even more dangerous than the version that NR says is too */
/* dangerous to use...No checking of bounds.  We might just run off */
/* to infinity... */
#define JMAX 20

static double
    rtnewt(void (*funcd)(double, double *, double *),
	   double x1, double xacc)
{
    int j;
    double df,dx,f,rtn;
    
    rtn=x1;
    for (j=1;j<=JMAX;j++) {
	(*funcd)(rtn,&f,&df);
	dx=f/df;
	Msg("rcrit", ("f(%g)=%g, dx=%g\n", rtn, f, -dx));
	rtn -= dx;
	if (fabs(dx) < xacc) return rtn;
    }
    Error("Maximum number of iterations exceeded in RTNEWT\n");
}

#undef JMAX
